import { Directive, Inject, Input, IterableDiffers, NgModule, Optional, } from '@angular/core';
import { coalesceWith } from '@rx-angular/cdk/coalescing';
import { combineLatest, merge, Observable, ReplaySubject, Subject, } from 'rxjs';
import { distinctUntilChanged, filter, map, startWith, switchMap, takeUntil, tap, } from 'rxjs/operators';
import { RxVirtualScrollStrategy, } from '../model';
import { unpatchedAnimationFrameTick, unpatchedMicroTask } from '../util';
import { DEFAULT_ITEM_SIZE, DEFAULT_RUNWAY_ITEMS, DEFAULT_RUNWAY_ITEMS_OPPOSITE, RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, } from '../virtual-scroll.config';
import * as i0 from "@angular/core";
/** @internal */
function removeFromArray(arr, index) {
    // perf: array.pop is faster than array.splice!
    if (index >= arr.length - 1) {
        return arr.pop();
    }
    else {
        return arr.splice(index, 1)[0];
    }
}
const defaultSizeExtract = (entry) => entry.borderBoxSize[0].blockSize;
/**
 * @Directive AutosizeVirtualScrollStrategy
 *
 * @description
 *
 * The `AutosizeVirtualScrollStrategy` provides a twitter-like virtual-scrolling
 * experience. It is able to render and position items based on their individual
 * size. It is comparable to \@angular/cdk/experimental `AutosizeVirtualScrollStrategy`, but
 * with a high performant layouting technique and more features.
 *
 * On top of this the `AutosizeVirtualScrollStrategy` is leveraging the native
 * `ResizeObserver` in order to detect size changes for each individual view
 * rendered to the DOM and properly re-position accordingly.
 *
 * In order to provide top-notch runtime performance the `AutosizeVirtualScrollStrategy`
 * builds up caches that prevent DOM interactions whenever possible. Once a view
 * was visited, its properties will be stored instead of re-read from the DOM again as
 * this can potentially lead to unwanted forced reflows.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
export class AutosizeVirtualScrollStrategy extends RxVirtualScrollStrategy {
    /** @internal */
    constructor(differs, defaults) {
        var _a, _b, _c, _d, _e, _f;
        super();
        this.differs = differs;
        this.defaults = defaults;
        /**
         * @description
         * The amount of items to render upfront in scroll direction
         */
        this.runwayItems = (_b = (_a = this.defaults) === null || _a === void 0 ? void 0 : _a.runwayItems) !== null && _b !== void 0 ? _b : DEFAULT_RUNWAY_ITEMS;
        /**
         * @description
         * The amount of items to render upfront in reverse scroll direction
         */
        this.runwayItemsOpposite = (_d = (_c = this.defaults) === null || _c === void 0 ? void 0 : _c.runwayItemsOpposite) !== null && _d !== void 0 ? _d : DEFAULT_RUNWAY_ITEMS_OPPOSITE;
        /**
         * @description
         * The default size of the items being rendered. The autosized strategy will assume
         * this size for items it doesn't know yet. For the smoothest experience,
         * you provide the mean size of all items being rendered - if possible of course.
         *
         * As soon as rxVirtualFor is able to also render actual tombstone items, this
         * will be the size of a tombstone item being rendered before the actual item
         * is inserted into its position.
         */
        this.tombstoneSize = (_f = (_e = this.defaults) === null || _e === void 0 ? void 0 : _e.itemSize) !== null && _f !== void 0 ? _f : DEFAULT_ITEM_SIZE;
        /** @internal */
        this.viewport = null;
        /** @internal */
        this.viewRepeater = null;
        /** @internal */
        this.dataDiffer = null;
        /** @internal */
        this._contentSize$ = new ReplaySubject(1);
        /** @internal */
        this.contentSize$ = this._contentSize$.asObservable();
        /** @internal */
        this._contentSize = 0;
        /** @internal */
        this._renderedRange$ = new ReplaySubject(1);
        /** @internal */
        this.renderedRange$ = this._renderedRange$.asObservable();
        /** @internal */
        this._renderedRange = { start: 0, end: 0 };
        // range of items where size is known and doesn't need to be re-calculated
        /** @internal */
        this._cachedSet = new Set();
        /** @internal */
        this._scrolledIndex$ = new ReplaySubject(1);
        /** @internal */
        this.scrolledIndex$ = this._scrolledIndex$.pipe(distinctUntilChanged());
        /** @internal */
        this._scrolledIndex = 0;
        /**
         * is set, when scrollToIndex is called and the scroll event touches
         * the edge of the viewport at the bottom. This is the only event where
         * we actually have to `guess` the scroll anchor and might end up having
         * to adjust
         * @internal
         * */
        this._scrollToEdgeIndex = null;
        /**
         * is set, when scrollToIndex is called
         * @internal
         * */
        this._scrollToAnchor = null;
        /** @internal */
        this.containerSize = 0;
        /** @internal */
        this.contentLength = 0;
        /** @internal */
        this._virtualItems = [];
        /** @internal */
        this.scrollTop = 0;
        /** @internal */
        this.direction = 'down';
        /** @internal */
        this.anchorScrollTop = 0;
        /** @internal */
        this.anchorItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.lastScreenItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.detached$ = new Subject();
        /** @internal */
        this.viewsResized$ = new Subject();
        /** @internal */
        this.afterViewResized$ = new Subject();
        /** @internal */
        this.runwayStateChanged$ = new Subject();
    }
    /** @internal */
    set contentSize(size) {
        this._contentSize = size;
        this._contentSize$.next(size);
    }
    /** @internal */
    set renderedRange(range) {
        this._renderedRange = range;
        this._renderedRange$.next(range);
    }
    /** @internal */
    get renderedRange() {
        return this._renderedRange;
    }
    /** @internal */
    get scrolledIndex() {
        return this._scrolledIndex;
    }
    /** @internal */
    set scrolledIndex(index) {
        this._scrolledIndex = index;
        this._scrolledIndex$.next(index);
    }
    /** @internal */
    until$() {
        return (o$) => o$.pipe(takeUntil(this.detached$));
    }
    get extractSize() {
        var _a, _b;
        return (_b = (_a = this.resizeObserverConfig) === null || _a === void 0 ? void 0 : _a.extractSize) !== null && _b !== void 0 ? _b : defaultSizeExtract;
    }
    /** @internal */
    ngOnChanges(changes) {
        if ((changes['runwayItemsOpposite'] &&
            !changes['runwayItemsOpposite'].firstChange) ||
            (changes['runwayItems'] && !changes['runwayItems'].firstChange)) {
            this.runwayStateChanged$.next();
        }
    }
    /** @internal */
    ngOnDestroy() {
        this.detach();
    }
    /** @internal */
    attach(viewport, viewRepeater) {
        this.viewport = viewport;
        this.viewRepeater = viewRepeater;
        this.resizeObserver = new ResizeObserver((events) => {
            this.viewsResized$.next(events);
        });
        this.maintainVirtualItems();
        this.calcRenderedRange();
        this.positionElements();
    }
    /** @internal */
    detach() {
        var _a;
        this.viewport = null;
        this.viewRepeater = null;
        this._virtualItems = [];
        (_a = this.resizeObserver) === null || _a === void 0 ? void 0 : _a.disconnect();
        this.resizeObserver = undefined;
        this.detached$.next();
    }
    scrollToIndex(index, behavior) {
        const _index = Math.min(Math.max(index, 0), this.contentLength - 1);
        let size = 0;
        for (let i = 0; i < _index; i++) {
            size += this.getItemSize(i);
        }
        this._scrollToAnchor = {
            index: _index,
            offset: 0,
        };
        this._scrollToEdgeIndex = null;
        if (size + this.containerSize > this._contentSize) {
            this._scrollToEdgeIndex = _index;
            this._scrollToAnchor = this.calculateAnchoredItem(this._scrollToAnchor, -Math.max(size + this.containerSize - this._contentSize));
        }
        this.viewport.scrollTo(size, behavior);
    }
    /** @internal */
    maintainVirtualItems() {
        // reset virtual viewport when opposite orientation to the scroll direction
        // changes, as we have to expect dimension changes for all items when this
        // happens. This could also be configurable as it maybe costs performance
        this.viewport.containerRect$.pipe(filter(() => this._virtualItems.length > 0), map(({ width }) => width), distinctUntilChanged(), this.until$()).subscribe(() => {
            // reset because we have no idea how items will behave
            let i = 0;
            let size = 0;
            this._cachedSet.clear();
            while (i < this.renderedRange.start) {
                this._virtualItems[i].size = 0;
                i++;
                size += this.tombstoneSize;
            }
            while (i >= this.renderedRange.start && i < this.renderedRange.end) {
                size += this._virtualItems[i].size;
                this._cachedSet.add(i);
                i++;
            }
            while (i < this.contentLength - 1) {
                this._virtualItems[i].size = 0;
                i++;
                size += this.tombstoneSize;
            }
            this.contentSize = size;
        });
        this.viewRepeater.values$.pipe(this.until$()).subscribe((items) => {
            var _a;
            const changes = (_a = this.getDiffer(items)) === null || _a === void 0 ? void 0 : _a.diff(items);
            if (changes) {
                // reset cache on update
                changes.forEachOperation((item, adjustedPreviousIndex, currentIndex) => {
                    if (item.previousIndex == null) {
                        const entry = {
                            size: 0,
                        };
                        if (currentIndex !== null &&
                            currentIndex < this._virtualItems.length) {
                            this._virtualItems.splice(currentIndex, 0, entry);
                        }
                        else {
                            this._virtualItems.push(entry);
                        }
                    }
                    else if (currentIndex === null) {
                        const removeIdx = adjustedPreviousIndex === null
                            ? this._virtualItems.length - 1
                            : adjustedPreviousIndex;
                        removeFromArray(this._virtualItems, removeIdx);
                        this._cachedSet.delete(removeIdx);
                    }
                    else if (adjustedPreviousIndex !== null) {
                        this._virtualItems[currentIndex] =
                            this._virtualItems[adjustedPreviousIndex];
                    }
                });
                changes.forEachIdentityChange(({ currentIndex }) => currentIndex && this._cachedSet.delete(currentIndex));
            }
            if (!this._contentSize && this._virtualItems.length > 0) {
                this.contentSize = this._virtualItems.length * this.tombstoneSize;
            }
        });
    }
    /** @internal */
    calcRenderedRange() {
        const dataLengthChanged$ = this.viewRepeater.values$.pipe(map((values) => (Array.isArray(values)
            ? values
            : values != null
                ? Array.from(values)
                : []).length), distinctUntilChanged());
        const onScroll$ = this.viewport.elementScrolled$.pipe(coalesceWith(unpatchedAnimationFrameTick()), map(() => this.viewport.getScrollTop()), startWith(0), tap((scrollTop) => {
            this.direction = scrollTop > this.scrollTop ? 'down' : 'up';
            this.scrollTop = scrollTop;
        }));
        combineLatest([
            dataLengthChanged$.pipe(tap((length) => {
                this.contentLength = length;
            })),
            this.viewport.containerRect$.pipe(map(({ height }) => height), distinctUntilChanged()),
            onScroll$,
            this.afterViewResized$.pipe(startWith(void 0)),
            this.runwayStateChanged$.pipe(startWith(void 0)),
        ])
            .pipe(map(([length, containerSize]) => {
            this.containerSize = containerSize;
            const range = { start: 0, end: 0 };
            const delta = this.scrollTop - this.anchorScrollTop;
            if (this._scrollToAnchor) {
                this.anchorItem = this._scrollToAnchor;
                this._scrollToAnchor = null;
            }
            else {
                // reset scrollToIndex after consequent scroll or any other update
                if (this._scrollToEdgeIndex !== null) {
                    this._scrollToEdgeIndex = null;
                }
                if (this.scrollTop == 0) {
                    this.anchorItem = { index: 0, offset: 0 };
                }
                else {
                    this.anchorItem = this.calculateAnchoredItem(this.anchorItem, delta);
                }
            }
            this.scrolledIndex = this.anchorItem.index;
            this.anchorScrollTop = this.scrollTop;
            this.lastScreenItem = this.calculateAnchoredItem(this.anchorItem, containerSize);
            if (this.direction === 'up') {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItems);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItemsOpposite);
            }
            else {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItemsOpposite);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItems);
            }
            return range;
        }))
            .pipe(distinctUntilChanged(({ start: prevStart, end: prevEnd }, { start, end }) => prevStart === start && prevEnd === end), this.until$())
            .subscribe((range) => (this.renderedRange = range));
    }
    /** @internal */
    positionElements() {
        this.viewRepeater.renderingStart$.pipe(switchMap(() => {
            const renderedRange = this.renderedRange;
            const adjustIndexWith = renderedRange.start;
            let scrolledIndex = null;
            let position = 0;
            let renderedViews = [];
            return merge(this.viewRepeater.viewRendered$.pipe(tap(({ view, index: viewIndex, item }) => {
                const adjustedIndex = viewIndex + adjustIndexWith;
                // this most of the time causes a forced reflow per rendered view.
                // it doesn't sound good, but it's still way more stable than
                // having one large reflow in a microtask after the actual
                // scheduler tick.
                // Right here, we can insert work into the task which is currently
                // executed as part of the concurrent scheduler tick.
                // causing the forced reflow here, also makes it count for the
                // schedulers frame budget. This way we will always run with the
                // configured FPS. The only case where this is not true is when rendering 1 single view
                // already explodes the budget
                this.updateElementSize(view, adjustedIndex);
                renderedViews.push({
                    view,
                    index: adjustedIndex,
                    item,
                });
            }), 
            // coalescing the layout to the next microtask
            // -> layout happens in the same tick as the scheduler. it will
            // be executed at the end and causes a forced reflow
            // to not have a reflow, we would need to let the browser stabilize the
            // current layout before reading `offsetHeight` of the nodes.
            // while the flames get smoother, other issues arise with that approach.
            // but here is def. room for improvement
            coalesceWith(unpatchedMicroTask()), map(() => {
                if (position === 0) {
                    position = this.calcInitialPosition(renderedRange);
                }
                let lastIndex = 0;
                let scrollTo = this.anchorScrollTop;
                // update DOM happens here
                // layout should be stable by now, we are thrashing it again, though
                for (let i = 0; i < renderedViews.length; i++) {
                    const { view, index, item } = renderedViews[i];
                    const element = this.getElement(view);
                    this.positionElement(element, position);
                    if (index === this._scrollToEdgeIndex) {
                        scrollTo = position;
                        this._scrollToEdgeIndex = null;
                    }
                    position += this._virtualItems[index].size;
                    if (scrolledIndex == null && position > this.scrollTop) {
                        scrolledIndex = index;
                    }
                    lastIndex = index;
                    this.viewRenderCallback.next({
                        index,
                        view,
                        item,
                    });
                }
                renderedViews = [];
                // this condition only covers the edge case when we scroll to the
                // lower edge of the viewport and we have to wait until the whole
                // renderedRange actually got rendered
                if (this._scrollToEdgeIndex === null) {
                    this.contentSize =
                        position + this.getRemainingSizeFrom(lastIndex + 1);
                    if (scrolledIndex != null &&
                        scrolledIndex !== this.scrolledIndex) {
                        this.scrolledIndex = scrolledIndex;
                    }
                    if (scrollTo !== this.scrollTop) {
                        const maxScrollTo = this._contentSize - this.containerSize;
                        if (scrollTo >= maxScrollTo &&
                            Math.ceil(this.scrollTop) >= maxScrollTo) {
                            // just trigger re-calculation of the renderedRange in case this happens
                            this.afterViewResized$.next();
                        }
                        else {
                            this.viewport.scrollTo(scrollTo);
                        }
                    }
                    else if (position < this.containerSize &&
                        lastIndex < this.contentLength) {
                        this.afterViewResized$.next();
                    }
                }
            })), this.viewRepeater.viewsRendered$.pipe(switchMap((views) => this.observeViewSizes$(adjustIndexWith, views).pipe(tap((lowestId) => {
                let i = lowestId;
                const range = {
                    start: i + adjustIndexWith,
                    end: renderedRange.end,
                };
                let position = this.calcInitialPosition(range);
                let index = range.start;
                for (i; i < views.length; i++) {
                    const element = this.getElement(views[i]);
                    this.positionElement(element, position);
                    index = i + adjustIndexWith;
                    position += this.getItemSize(index);
                }
                this.contentSize =
                    position + this.getRemainingSizeFrom(index + 1);
                this.afterViewResized$.next();
            })))));
        }), this.until$()).subscribe();
    }
    /**
     * @internal
     * heavily inspired by
     *   https://github.com/GoogleChromeLabs/ui-element-samples/blob/gh-pages/infinite-scroller/scripts/infinite-scroll.js
     */
    calculateAnchoredItem(initialAnchor, delta) {
        var _a;
        if (delta == 0)
            return initialAnchor;
        delta += initialAnchor.offset;
        let i = initialAnchor.index;
        let tombstones = 0;
        const items = this._virtualItems;
        if (delta < 0) {
            while (delta < 0 && i > 0 && items[i - 1].size) {
                delta += items[i - 1].size;
                i--;
            }
            tombstones = Math.max(-i, Math.ceil(Math.min(delta, 0) / this.tombstoneSize));
        }
        else {
            while (delta > 0 &&
                i < this.contentLength &&
                ((_a = items[i]) === null || _a === void 0 ? void 0 : _a.size) &&
                items[i].size < delta) {
                delta -= items[i].size;
                i++;
            }
            if (i >= this.contentLength) {
                tombstones = 0;
            }
            else if (!items[i].size) {
                tombstones = Math.floor(Math.max(delta, 0) / this.tombstoneSize);
            }
        }
        i += tombstones;
        delta -= tombstones * this.tombstoneSize;
        return {
            index: Math.min(i, this.contentLength),
            offset: Math.max(0, delta),
        };
    }
    /** @internal */
    calcInitialPosition(range) {
        let pos = 0;
        let i = 0;
        this.anchorScrollTop = 0;
        for (i = 0; i < this.anchorItem.index; i++) {
            this.anchorScrollTop += this.getItemSize(i);
        }
        this.anchorScrollTop += this.anchorItem.offset;
        // Calculate position of starting node
        pos = this.anchorScrollTop - this.anchorItem.offset;
        i = this.anchorItem.index;
        while (i > range.start) {
            const itemSize = this.getItemSize(i - 1);
            pos -= itemSize;
            i--;
        }
        while (i < range.start) {
            const itemSize = this.getItemSize(i);
            pos += itemSize;
            i++;
        }
        return pos;
    }
    /** @internal */
    updateElementSize(view, index) {
        const oldSize = this._virtualItems[index].size;
        const isCached = this._cachedSet.has(index) && oldSize > 0;
        const size = isCached
            ? oldSize
            : this.getElementSize(this.getElement(view));
        this._virtualItems[index] = { size };
        this._cachedSet.add(index);
        return size;
    }
    /** @internal */
    observeViewSizes$(adjustIndexWith, views) {
        var _a;
        const elementCache = new WeakMap();
        let lowestResizedId;
        for (let i = 0; i < views.length; i++) {
            const element = this.getElement(views[i]);
            this.resizeObserver.observe(element, (_a = this.resizeObserverConfig) === null || _a === void 0 ? void 0 : _a.options);
            elementCache.set(element, i);
        }
        return new Observable((observer) => {
            const inner = this.viewsResized$.subscribe((events) => {
                events.forEach((event) => {
                    if (!event.target.isConnected) {
                        return;
                    }
                    const cachedId = elementCache.get(event.target);
                    if (cachedId !== undefined) {
                        const adjustedId = cachedId + adjustIndexWith;
                        const size = Math.round(this.extractSize(event));
                        if (this._virtualItems[adjustedId].size !== size) {
                            lowestResizedId = Math.min(lowestResizedId !== null && lowestResizedId !== void 0 ? lowestResizedId : Number.MAX_SAFE_INTEGER, cachedId);
                            this._virtualItems[adjustedId].size = size;
                        }
                    }
                });
                if (lowestResizedId !== undefined) {
                    observer.next(lowestResizedId);
                }
            });
            return () => {
                var _a;
                for (let i = 0; i < views.length; i++) {
                    const element = this.getElement(views[i]);
                    (_a = this.resizeObserver) === null || _a === void 0 ? void 0 : _a.unobserve(element);
                }
                inner.unsubscribe();
            };
        }).pipe(coalesceWith(unpatchedAnimationFrameTick()), map((lowestId) => {
            lowestResizedId = undefined;
            return lowestId;
        }));
    }
    /** @internal */
    getRemainingSizeFrom(from) {
        let remaining = 0;
        for (let i = from; i < this.contentLength; i++) {
            remaining += this.getItemSize(i);
        }
        return remaining;
    }
    /** @internal */
    getItemSize(index) {
        return this._virtualItems[index].size || this.tombstoneSize;
    }
    /** @internal */
    getElementSize(element) {
        return element.offsetHeight;
    }
    /** @internal */
    positionElement(element, scrollTop) {
        element.style.position = 'absolute';
        element.style.transform = `translateY(${scrollTop}px)`;
    }
    /** @internal */
    getDiffer(values) {
        if (this.dataDiffer) {
            return this.dataDiffer;
        }
        return values
            ? (this.dataDiffer = this.differs
                .find(values)
                .create(this.viewRepeater._trackBy))
            : null;
    }
}
/** @nocollapse */ AutosizeVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategy, deps: [{ token: i0.IterableDiffers }, { token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ AutosizeVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: AutosizeVirtualScrollStrategy, selector: "rx-virtual-scroll-viewport[autosize]", inputs: { runwayItems: "runwayItems", runwayItemsOpposite: "runwayItemsOpposite", tombstoneSize: "tombstoneSize", resizeObserverConfig: "resizeObserverConfig" }, providers: [
        {
            provide: RxVirtualScrollStrategy,
            useExisting: AutosizeVirtualScrollStrategy,
        },
    ], usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategy, decorators: [{
            type: Directive,
            args: [{
                    // eslint-disable-next-line @angular-eslint/directive-selector
                    selector: 'rx-virtual-scroll-viewport[autosize]',
                    providers: [
                        {
                            provide: RxVirtualScrollStrategy,
                            useExisting: AutosizeVirtualScrollStrategy,
                        },
                    ],
                }]
        }], ctorParameters: function () { return [{ type: i0.IterableDiffers }, { type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { runwayItems: [{
                type: Input
            }], runwayItemsOpposite: [{
                type: Input
            }], tombstoneSize: [{
                type: Input
            }], resizeObserverConfig: [{
                type: Input
            }] } });
export class AutosizeVirtualScrollStrategyModule {
}
/** @nocollapse */ AutosizeVirtualScrollStrategyModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ AutosizeVirtualScrollStrategyModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, declarations: [AutosizeVirtualScrollStrategy], exports: [AutosizeVirtualScrollStrategy] });
/** @nocollapse */ AutosizeVirtualScrollStrategyModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: AutosizeVirtualScrollStrategyModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [AutosizeVirtualScrollStrategy],
                    declarations: [AutosizeVirtualScrollStrategy],
                    providers: [],
                }]
        }] });
//# sourceMappingURL=autosized-virtual-scroll-strategy.js.map