import { Directive, Inject, Input, IterableDiffers, NgModule, Optional, } from '@angular/core';
import { combineLatest, ReplaySubject, Subject, } from 'rxjs';
import { debounce, distinctUntilChanged, map, startWith, switchMap, takeUntil, tap, } from 'rxjs/operators';
import { RxVirtualScrollStrategy, } from '../model';
import { unpatchedAnimationFrameTick } from '../util';
import { DEFAULT_ITEM_SIZE, DEFAULT_RUNWAY_ITEMS, DEFAULT_RUNWAY_ITEMS_OPPOSITE, RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, } from '../virtual-scroll.config';
import * as i0 from "@angular/core";
/** @internal */
function removeFromArray(arr, index) {
    // perf: array.pop is faster than array.splice!
    if (index >= arr.length - 1) {
        return arr.pop();
    }
    else {
        return arr.splice(index, 1)[0];
    }
}
const defaultItemSize = () => DEFAULT_ITEM_SIZE;
/**
 * @Directive DynamicSizeVirtualScrollStrategy
 *
 * @description
 *
 * The `DynamicSizeVirtualScrollStrategy` is very similar to the `AutosizeVirtualScrollStrategy`.
 * It positions items based on a function determining its size.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
export class DynamicSizeVirtualScrollStrategy extends RxVirtualScrollStrategy {
    /** @internal */
    constructor(differs, defaults) {
        var _a, _b, _c, _d;
        super();
        this.differs = differs;
        this.defaults = defaults;
        /**
         * @description
         * The amount of items to render upfront in scroll direction
         */
        this.runwayItems = (_b = (_a = this.defaults) === null || _a === void 0 ? void 0 : _a.runwayItems) !== null && _b !== void 0 ? _b : DEFAULT_RUNWAY_ITEMS;
        /**
         * @description
         * The amount of items to render upfront in reverse scroll direction
         */
        this.runwayItemsOpposite = (_d = (_c = this.defaults) === null || _c === void 0 ? void 0 : _c.runwayItemsOpposite) !== null && _d !== void 0 ? _d : DEFAULT_RUNWAY_ITEMS_OPPOSITE;
        this._itemSizeFn = defaultItemSize;
        /** @internal */
        this.viewport = null;
        /** @internal */
        this.viewRepeater = null;
        /** @internal */
        this.dataDiffer = null;
        /** @internal */
        this._contentSize$ = new ReplaySubject(1);
        /** @internal */
        this.contentSize$ = this._contentSize$.asObservable();
        /** @internal */
        this._contentSize = 0;
        /** @internal */
        this._renderedRange$ = new ReplaySubject(1);
        /** @internal */
        this.renderedRange$ = this._renderedRange$.asObservable();
        /** @internal */
        this._renderedRange = { start: 0, end: 0 };
        /** @internal */
        this._scrolledIndex$ = new ReplaySubject(1);
        /** @internal */
        this.scrolledIndex$ = this._scrolledIndex$.pipe(distinctUntilChanged());
        /** @internal */
        this.contentLength = 0;
        /** @internal */
        this._virtualItems = [];
        /** @internal */
        this.scrollTop = 0;
        /** @internal */
        this.direction = 'down';
        /** @internal */
        this.anchorScrollTop = 0;
        /** @internal */
        this.anchorItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.lastScreenItem = {
            index: 0,
            offset: 0,
        };
        /** @internal */
        this.detached$ = new Subject();
        /** @internal */
        this.runwayStateChanged$ = new Subject();
    }
    /**
     * @description
     * Function returning the size of an item
     */
    set itemSize(fn) {
        if (fn) {
            this._itemSizeFn = fn;
        }
    }
    get itemSize() {
        return this._itemSizeFn;
    }
    /** @internal */
    set contentSize(size) {
        this._contentSize = size;
        this._contentSize$.next(size);
    }
    // range of items where size is known and doesn't need to be re-calculated
    /** @internal */
    set renderedRange(range) {
        this._renderedRange = range;
        this._renderedRange$.next(range);
    }
    /** @internal */
    get renderedRange() {
        return this._renderedRange;
    }
    /** @internal */
    set scrolledIndex(index) {
        this._scrolledIndex$.next(index);
    }
    /** @internal */
    until$() {
        return (o$) => o$.pipe(takeUntil(this.detached$));
    }
    /** @internal */
    ngOnChanges(changes) {
        if ((changes['runwayItemsOpposite'] &&
            !changes['runwayItemsOpposite'].firstChange) ||
            (changes['runwayItems'] && !changes['runwayItems'].firstChange)) {
            this.runwayStateChanged$.next();
        }
    }
    /** @internal */
    ngOnDestroy() {
        this.detach();
    }
    /** @internal */
    attach(viewport, viewRepeater) {
        this.viewport = viewport;
        this.viewRepeater = viewRepeater;
        this.maintainVirtualItems();
        this.calcRenderedRange();
        this.positionElements();
    }
    /** @internal */
    detach() {
        this.viewport = null;
        this.viewRepeater = null;
        this._virtualItems = [];
        this.detached$.next();
    }
    scrollToIndex(index, behavior) {
        const _index = Math.min(Math.max(index, 0), this.contentLength - 1);
        let offset = 0;
        for (let i = 0; i < _index; i++) {
            offset += this._virtualItems[i].size;
        }
        this.viewport.scrollTo(offset, behavior);
    }
    /** @internal */
    maintainVirtualItems() {
        this.viewRepeater.values$.pipe(this.until$()).subscribe((items) => {
            var _a;
            const changes = (_a = this.getDiffer(items)) === null || _a === void 0 ? void 0 : _a.diff(items);
            if (changes) {
                changes.forEachOperation((change, adjustedPreviousIndex, currentIndex) => {
                    if (change.previousIndex == null) {
                        const entry = {
                            size: this.itemSize(change.item),
                        };
                        if (currentIndex !== null &&
                            currentIndex < this._virtualItems.length) {
                            this._virtualItems.splice(currentIndex, 0, entry);
                        }
                        else {
                            this._virtualItems.push(entry);
                        }
                    }
                    else if (currentIndex == null) {
                        const removeIdx = adjustedPreviousIndex == null
                            ? this._virtualItems.length - 1
                            : adjustedPreviousIndex;
                        removeFromArray(this._virtualItems, removeIdx);
                    }
                    else if (adjustedPreviousIndex !== null) {
                        this._virtualItems[currentIndex] =
                            this._virtualItems[adjustedPreviousIndex];
                    }
                });
                changes.forEachIdentityChange((record) => {
                    this._virtualItems[record.currentIndex] = {
                        size: this.itemSize(record.item),
                    };
                });
                let contentSize = 0;
                for (let i = 0; i < this._virtualItems.length; i++) {
                    contentSize += this._virtualItems[i].size;
                }
                this.contentSize = contentSize;
            }
        });
    }
    /** @internal */
    calcRenderedRange() {
        const dataLengthChanged$ = this.viewRepeater.values$.pipe(map((values) => (Array.isArray(values)
            ? values
            : values != null
                ? Array.from(values)
                : []).length), distinctUntilChanged());
        const onScroll$ = this.viewport.elementScrolled$.pipe(debounce(() => unpatchedAnimationFrameTick()), map(() => this.viewport.getScrollTop()), startWith(0), tap((_scrollTop) => {
            this.direction = _scrollTop > this.scrollTop ? 'down' : 'up';
            this.scrollTop = _scrollTop;
        }));
        combineLatest([
            dataLengthChanged$.pipe(tap((length) => {
                this.contentLength = length;
            })),
            this.viewport.containerRect$.pipe(map(({ height }) => height), distinctUntilChanged()),
            onScroll$,
            this.runwayStateChanged$.pipe(startWith(void 0)),
        ])
            .pipe(map(([length, containerHeight]) => {
            const range = { start: 0, end: 0 };
            const delta = this.scrollTop - this.anchorScrollTop;
            if (this.scrollTop == 0) {
                this.anchorItem = { index: 0, offset: 0 };
            }
            else {
                this.anchorItem = this.calculateAnchoredItem(this.anchorItem, delta);
            }
            this.scrolledIndex = this.anchorItem.index;
            this.anchorScrollTop = this.scrollTop;
            this.lastScreenItem = this.calculateAnchoredItem(this.anchorItem, containerHeight);
            if (this.direction === 'up') {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItems);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItemsOpposite);
            }
            else {
                range.start = Math.max(0, this.anchorItem.index - this.runwayItemsOpposite);
                range.end = Math.min(length, this.lastScreenItem.index + this.runwayItems);
            }
            return range;
        }))
            .pipe(distinctUntilChanged(({ start: prevStart, end: prevEnd }, { start, end }) => prevStart === start && prevEnd === end), this.until$())
            .subscribe((range) => (this.renderedRange = range));
    }
    /** @internal */
    positionElements() {
        this.viewRepeater.renderingStart$.pipe(switchMap(() => {
            const renderedRange = this.renderedRange;
            const adjustIndexWith = renderedRange.start;
            let scrolledIndex = null;
            let position = 0;
            return this.viewRepeater.viewRendered$.pipe(map(({ view, index: viewIndex, item }, idx) => {
                const index = viewIndex + adjustIndexWith;
                if (idx === 0) {
                    position = this.calcInitialPosition(renderedRange);
                }
                const size = this.getItemSize(index);
                this.positionElement(this.getElement(view), position);
                position += size;
                if (scrolledIndex == null && position > this.scrollTop) {
                    scrolledIndex = index;
                }
                if (scrolledIndex != null) {
                    this.scrolledIndex = scrolledIndex;
                }
                this.viewRenderCallback.next({
                    index,
                    view,
                    item,
                });
            }));
        }), this.until$()).subscribe();
    }
    /**
     * @internal
     * heavily inspired by
     *   https://github.com/GoogleChromeLabs/ui-element-samples/blob/gh-pages/infinite-scroller/scripts/infinite-scroll.js
     */
    calculateAnchoredItem(initialAnchor, delta) {
        if (delta == 0)
            return initialAnchor;
        delta += initialAnchor.offset;
        let i = initialAnchor.index;
        const items = this._virtualItems;
        if (delta < 0) {
            while (delta < 0 && i > 0) {
                delta += items[i - 1].size;
                i--;
            }
        }
        else {
            while (delta > 0 && i < this.contentLength && items[i].size < delta) {
                delta -= items[i].size;
                i++;
            }
        }
        return {
            index: Math.min(i, this.contentLength),
            offset: delta,
        };
    }
    /** @internal */
    calcInitialPosition(range) {
        let pos = 0;
        let i = 0;
        this.anchorScrollTop = 0;
        for (i = 0; i < this.anchorItem.index; i++) {
            this.anchorScrollTop += this.getItemSize(i);
        }
        this.anchorScrollTop += this.anchorItem.offset;
        // Calculate position of starting node
        pos = this.anchorScrollTop - this.anchorItem.offset;
        i = this.anchorItem.index;
        while (i > range.start) {
            const itemSize = this.getItemSize(i - 1);
            pos -= itemSize;
            i--;
        }
        while (i < range.start) {
            const itemSize = this.getItemSize(i);
            pos += itemSize;
            i++;
        }
        return pos;
    }
    /** @internal */
    getItemSize(index) {
        return this._virtualItems[index].size;
    }
    /** @internal */
    positionElement(element, scrollTop) {
        element.style.position = 'absolute';
        element.style.transform = `translateY(${scrollTop}px)`;
    }
    /** @internal */
    getDiffer(values) {
        if (this.dataDiffer) {
            return this.dataDiffer;
        }
        return values
            ? (this.dataDiffer = this.differs
                .find(values)
                .create(this.viewRepeater._trackBy))
            : null;
    }
}
/** @nocollapse */ DynamicSizeVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategy, deps: [{ token: i0.IterableDiffers }, { token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ DynamicSizeVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: DynamicSizeVirtualScrollStrategy, selector: "rx-virtual-scroll-viewport[dynamic]", inputs: { runwayItems: "runwayItems", runwayItemsOpposite: "runwayItemsOpposite", itemSize: ["dynamic", "itemSize"] }, providers: [
        {
            provide: RxVirtualScrollStrategy,
            useExisting: DynamicSizeVirtualScrollStrategy,
        },
    ], usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategy, decorators: [{
            type: Directive,
            args: [{
                    // eslint-disable-next-line @angular-eslint/directive-selector
                    selector: 'rx-virtual-scroll-viewport[dynamic]',
                    providers: [
                        {
                            provide: RxVirtualScrollStrategy,
                            useExisting: DynamicSizeVirtualScrollStrategy,
                        },
                    ],
                }]
        }], ctorParameters: function () { return [{ type: i0.IterableDiffers }, { type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { runwayItems: [{
                type: Input
            }], runwayItemsOpposite: [{
                type: Input
            }], itemSize: [{
                type: Input,
                args: ['dynamic']
            }] } });
export class DynamicSizeVirtualScrollStrategyModule {
}
/** @nocollapse */ DynamicSizeVirtualScrollStrategyModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ DynamicSizeVirtualScrollStrategyModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, declarations: [DynamicSizeVirtualScrollStrategy], exports: [DynamicSizeVirtualScrollStrategy] });
/** @nocollapse */ DynamicSizeVirtualScrollStrategyModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: DynamicSizeVirtualScrollStrategyModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [DynamicSizeVirtualScrollStrategy],
                    declarations: [DynamicSizeVirtualScrollStrategy],
                    providers: [],
                }]
        }] });
//# sourceMappingURL=dynamic-size-virtual-scroll-strategy.js.map