import { Directive, Inject, Input, NgModule, Optional, } from '@angular/core';
import { coalesceWith } from '@rx-angular/cdk/coalescing';
import { combineLatest, ReplaySubject, Subject, } from 'rxjs';
import { distinctUntilChanged, map, startWith, switchMap, takeUntil, tap, } from 'rxjs/operators';
import { RxVirtualScrollStrategy, } from '../model';
import { unpatchedAnimationFrameTick } from '../util';
import { DEFAULT_ITEM_SIZE, DEFAULT_RUNWAY_ITEMS, DEFAULT_RUNWAY_ITEMS_OPPOSITE, RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, } from '../virtual-scroll.config';
import * as i0 from "@angular/core";
/**
 * @Directive FixedSizeVirtualScrollStrategy
 *
 * @description
 *
 * The `FixedSizeVirtualScrollStrategy` provides a very performant way of rendering
 * items of a given size. It is comparable to \@angular/cdk `FixedSizeVirtualScrollStrategy`, but
 * with a high performant layouting technique.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
export class FixedSizeVirtualScrollStrategy extends RxVirtualScrollStrategy {
    constructor(defaults) {
        var _a, _b, _c, _d;
        super();
        this.defaults = defaults;
        this._itemSize = DEFAULT_ITEM_SIZE;
        /**
         * @description
         * The amount of items to render upfront in scroll direction
         */
        this.runwayItems = (_b = (_a = this.defaults) === null || _a === void 0 ? void 0 : _a.runwayItems) !== null && _b !== void 0 ? _b : DEFAULT_RUNWAY_ITEMS;
        /**
         * @description
         * The amount of items to render upfront in reverse scroll direction
         */
        this.runwayItemsOpposite = (_d = (_c = this.defaults) === null || _c === void 0 ? void 0 : _c.runwayItemsOpposite) !== null && _d !== void 0 ? _d : DEFAULT_RUNWAY_ITEMS_OPPOSITE;
        /** @internal */
        this.runwayStateChanged$ = new Subject();
        this.viewport = null;
        this.viewRepeater = null;
        this._scrolledIndex$ = new ReplaySubject(1);
        this.scrolledIndex$ = this._scrolledIndex$.pipe(distinctUntilChanged());
        this._scrolledIndex = 0;
        this._contentSize$ = new ReplaySubject(1);
        this.contentSize$ = this._contentSize$.asObservable();
        this._contentSize = 0;
        this._renderedRange$ = new ReplaySubject(1);
        this.renderedRange$ = this._renderedRange$.asObservable();
        this._renderedRange = { start: 0, end: 0 };
        this.scrollTop = 0;
        this.direction = 'down';
        this.detached$ = new Subject();
    }
    /**
     * @description
     * The size of the items in the virtually scrolled list
     */
    set itemSize(itemSize) {
        if (typeof itemSize === 'number') {
            this._itemSize = itemSize;
        }
    }
    get itemSize() {
        return this._itemSize;
    }
    set scrolledIndex(index) {
        this._scrolledIndex = index;
        this._scrolledIndex$.next(index);
    }
    set contentSize(size) {
        this._contentSize = size;
        this._contentSize$.next(size);
    }
    set renderedRange(range) {
        this._renderedRange = range;
        this._renderedRange$.next(range);
    }
    get renderedRange() {
        return this._renderedRange;
    }
    /** @internal */
    ngOnChanges(changes) {
        if ((changes['runwayItemsOpposite'] &&
            !changes['runwayItemsOpposite'].firstChange) ||
            (changes['runwayItems'] && !changes['runwayItems'].firstChange)) {
            this.runwayStateChanged$.next();
        }
    }
    ngOnDestroy() {
        this.detach();
    }
    attach(viewport, viewRepeater) {
        this.viewport = viewport;
        this.viewRepeater = viewRepeater;
        this.calcRenderedRange();
        this.positionElements();
    }
    detach() {
        this.viewport = null;
        this.viewRepeater = null;
        this.detached$.next();
    }
    positionElements() {
        this.viewRepeater.renderingStart$.pipe(switchMap(() => {
            const start = this.renderedRange.start;
            return this.viewRepeater.viewRendered$.pipe(tap(({ view, index, item }) => {
                this._setViewPosition(view, (index + start) * this.itemSize);
                this.viewRenderCallback.next({
                    view,
                    item,
                    index,
                });
            }));
        }), this.untilDetached$()).subscribe();
    }
    calcRenderedRange() {
        const dataLengthChanged$ = this.viewRepeater.values$.pipe(map((values) => (Array.isArray(values)
            ? values
            : values != null
                ? Array.from(values)
                : []).length), distinctUntilChanged(), tap((dataLength) => (this.contentSize = dataLength * this.itemSize)));
        const onScroll$ = this.viewport.elementScrolled$.pipe(coalesceWith(unpatchedAnimationFrameTick()), map(() => this.viewport.getScrollTop()), startWith(0), tap((_scrollTop) => {
            this.direction = _scrollTop > this.scrollTop ? 'down' : 'up';
            this.scrollTop = _scrollTop;
        }));
        combineLatest([
            dataLengthChanged$,
            this.viewport.containerRect$.pipe(map(({ height }) => height), distinctUntilChanged()),
            onScroll$,
            this.runwayStateChanged$.pipe(startWith(void 0)),
        ])
            .pipe(map(([length, containerSize]) => {
            const range = { start: 0, end: 0 };
            if (this.direction === 'up') {
                range.start = Math.floor(Math.max(0, this.scrollTop - this.runwayItems * this.itemSize) /
                    this.itemSize);
                range.end = Math.min(length, Math.ceil((this.scrollTop +
                    containerSize +
                    this.runwayItemsOpposite * this.itemSize) /
                    this.itemSize));
            }
            else {
                range.start = Math.floor(Math.max(0, this.scrollTop - this.runwayItemsOpposite * this.itemSize) / this.itemSize);
                range.end = Math.min(length, Math.ceil((this.scrollTop +
                    containerSize +
                    this.runwayItems * this.itemSize) /
                    this.itemSize));
            }
            this.scrolledIndex = Math.floor(this.scrollTop / this.itemSize);
            return range;
        }), distinctUntilChanged(({ start: prevStart, end: prevEnd }, { start, end }) => prevStart === start && prevEnd === end), this.untilDetached$())
            .subscribe((range) => (this.renderedRange = range));
    }
    scrollToIndex(index, behavior) {
        const scrollTop = this.itemSize * index;
        this.viewport.scrollTo(scrollTop, behavior);
    }
    untilDetached$() {
        return (o$) => o$.pipe(takeUntil(this.detached$));
    }
    _setViewPosition(view, scrollTop) {
        const element = this.getElement(view);
        element.style.position = 'absolute';
        element.style.transform = `translateY(${scrollTop}px)`;
    }
}
/** @nocollapse */ FixedSizeVirtualScrollStrategy.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategy, deps: [{ token: RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ FixedSizeVirtualScrollStrategy.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.16", type: FixedSizeVirtualScrollStrategy, selector: "rx-virtual-scroll-viewport[itemSize]", inputs: { itemSize: "itemSize", runwayItems: "runwayItems", runwayItemsOpposite: "runwayItemsOpposite" }, providers: [
        {
            provide: RxVirtualScrollStrategy,
            useExisting: FixedSizeVirtualScrollStrategy,
        },
    ], usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategy, decorators: [{
            type: Directive,
            args: [{
                    // eslint-disable-next-line @angular-eslint/directive-selector
                    selector: 'rx-virtual-scroll-viewport[itemSize]',
                    providers: [
                        {
                            provide: RxVirtualScrollStrategy,
                            useExisting: FixedSizeVirtualScrollStrategy,
                        },
                    ],
                }]
        }], ctorParameters: function () { return [{ type: undefined, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [RX_VIRTUAL_SCROLL_DEFAULT_OPTIONS]
                }] }]; }, propDecorators: { itemSize: [{
                type: Input
            }], runwayItems: [{
                type: Input
            }], runwayItemsOpposite: [{
                type: Input
            }] } });
export class FixedSizeVirtualScrollStrategyModule {
}
/** @nocollapse */ FixedSizeVirtualScrollStrategyModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ FixedSizeVirtualScrollStrategyModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, declarations: [FixedSizeVirtualScrollStrategy], exports: [FixedSizeVirtualScrollStrategy] });
/** @nocollapse */ FixedSizeVirtualScrollStrategyModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, providers: [], imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: FixedSizeVirtualScrollStrategyModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [],
                    exports: [FixedSizeVirtualScrollStrategy],
                    declarations: [FixedSizeVirtualScrollStrategy],
                    providers: [],
                }]
        }] });
//# sourceMappingURL=fixed-size-virtual-scroll-strategy.js.map