import { onStrategy, } from '@rx-angular/cdk/render-strategies';
import { Subject } from 'rxjs';
/**
 * @internal
 *
 * Factory that returns a `ListTemplateManager` for the passed params.
 *
 * @param templateSettings
 */
export function getVirtualTemplateHandler({ viewContainerRef, templateRef, createViewContext, updateViewContext, viewCacheSize, }) {
    let _viewCache = [];
    const _viewRendered$ = new Subject();
    return {
        toTemplateWork,
        viewRendered$: _viewRendered$,
        detach: () => {
            for (const view of _viewCache) {
                view.destroy();
            }
            _viewCache = [];
        },
    };
    function toTemplateWork(iterableChanges, items, renderedRange, strategy, count, ngZone) {
        const listChanges = getListChanges(iterableChanges, items);
        const changes = listChanges[0].sort(([, payloadA], [, payloadB]) => {
            return (payloadA[1] || 0) - (payloadB[1] || 0);
        });
        const insertedOrRemoved = listChanges[1];
        return {
            insertedOrRemoved,
            changes,
            work$: changes.length > 0
                ? changes.map((change) => {
                    return onStrategy(change, strategy, (_change) => {
                        const [type, [item, index, adjustedPreviousIndex]] = _change;
                        const update = () => ({
                            view: viewContainerRef.get(index || 0),
                            index,
                            change: type,
                            item,
                        });
                        switch (type) {
                            case 0 /* insert */:
                                _insertView(item, index, count, renderedRange.start + index);
                                // console.log('perform insert', update());
                                _viewRendered$.next(update());
                                break;
                            case 2 /* move */:
                                _moveView(item, adjustedPreviousIndex, index, count, renderedRange.start + index);
                                // console.log('perform move', update());
                                _viewRendered$.next(update());
                                break;
                            case 1 /* remove */:
                                if (index != null) {
                                    _detachAndCacheView(index);
                                }
                                // console.log('perform remove', update());
                                // _viewRendered$.next(update());
                                break;
                            case 3 /* update */:
                                _updateView(item, index, count, renderedRange.start + index);
                                // console.log('perform update', update());
                                _viewRendered$.next(update());
                                break;
                            case 4 /* context */:
                                _updateView(item, index, count, renderedRange.start + index);
                                // console.log('perform context', update());
                                _viewRendered$.next(update());
                                break;
                        }
                    }, { ngZone });
                })
                : [],
        };
    }
    function _updateUnchangedContext(index, count, contextIndex) {
        const view = viewContainerRef.get(index);
        view.context.updateContext({
            count,
            index: contextIndex,
        });
        view.detectChanges();
    }
    function _updateView(item, index, count, contextIndex) {
        const view = viewContainerRef.get(index);
        updateViewContext(item, view, {
            count,
            index: contextIndex,
        });
        view.detectChanges();
    }
    /**
     * Inserts a view for a new item, either from the cache or by creating a new
     * one. Returns `undefined` if the item was inserted into a cached view.
     */
    function _insertView(value, currentIndex, count, contextIndex) {
        const cachedView = _insertViewFromCache(currentIndex);
        if (cachedView) {
            updateViewContext(value, cachedView, {
                count,
                index: contextIndex,
            });
            cachedView.detectChanges();
            return undefined;
        }
        const context = createViewContext(value, {
            count,
            index: contextIndex,
        });
        const view = viewContainerRef.createEmbeddedView(templateRef, context, currentIndex);
        view.detectChanges();
        return view;
    }
    /** Detaches the view at the given index and inserts into the view cache. */
    function _detachAndCacheView(index) {
        const detachedView = viewContainerRef.detach(index);
        _maybeCacheView(detachedView);
        detachedView.detectChanges();
    }
    /** Moves view at the previous index to the current index. */
    function _moveView(value, adjustedPreviousIndex, currentIndex, count, contextIndex) {
        const view = viewContainerRef.get(adjustedPreviousIndex);
        viewContainerRef.move(view, currentIndex);
        updateViewContext(value, view, {
            count,
            index: contextIndex,
        });
        view.detectChanges();
        return view;
    }
    /**
     * Cache the given detached view. If the cache is full, the view will be
     * destroyed.
     */
    function _maybeCacheView(view) {
        if (_viewCache.length < viewCacheSize) {
            _viewCache.push(view);
            return true;
        }
        else {
            const index = viewContainerRef.indexOf(view);
            // The host component could remove views from the container outside of
            // the view repeater. It's unlikely this will occur, but just in case,
            // destroy the view on its own, otherwise destroy it through the
            // container to ensure that all the references are removed.
            if (index === -1) {
                view.destroy();
            }
            else {
                viewContainerRef.remove(index);
            }
            return false;
        }
    }
    /** Inserts a recycled view from the cache at the given index. */
    function _insertViewFromCache(index) {
        const cachedView = _viewCache.pop();
        if (cachedView) {
            viewContainerRef.insert(cachedView, index);
        }
        return cachedView || null;
    }
}
/**
 * @internal
 *
 * @param changes
 * @param items
 */
function getListChanges(changes, items) {
    const changedIdxs = new Set();
    const changesArr = [];
    let notifyParent = false;
    changes.forEachOperation(({ item, previousIndex }, adjustedPreviousIndex, currentIndex) => {
        if (previousIndex == null) {
            // insert
            changesArr.push(getInsertChange(item, currentIndex));
            changedIdxs.add(item);
            notifyParent = true;
        }
        else if (currentIndex == null) {
            // remove
            changesArr.push(getRemoveChange(item, adjustedPreviousIndex));
            notifyParent = true;
        }
        else if (adjustedPreviousIndex !== null) {
            // move
            changesArr.push(getMoveChange(item, currentIndex, adjustedPreviousIndex));
            changedIdxs.add(item);
            notifyParent = true;
        }
    });
    changes.forEachIdentityChange(({ item, currentIndex }) => {
        if (currentIndex != null && !changedIdxs.has(item)) {
            changesArr.push(getUpdateChange(item, currentIndex));
            changedIdxs.add(item);
        }
    });
    if (notifyParent) {
        for (let i = 0; i < items.length; i++) {
            const item = items[i];
            if (!changedIdxs.has(item)) {
                changesArr.push(getUnchangedChange(item, i));
            }
        }
    }
    return [changesArr, notifyParent];
    // ==========
    function getMoveChange(item, currentIndex, adjustedPreviousIndex) {
        return [
            2 /* move */,
            [item, currentIndex, adjustedPreviousIndex],
        ];
    }
    function getUpdateChange(item, currentIndex) {
        return [3 /* update */, [item, currentIndex]];
    }
    function getUnchangedChange(item, index) {
        return [4 /* context */, [item, index]];
    }
    function getInsertChange(item, currentIndex) {
        return [
            0 /* insert */,
            [item, currentIndex === null ? undefined : currentIndex],
        ];
    }
    function getRemoveChange(item, adjustedPreviousIndex) {
        return [
            1 /* remove */,
            [
                item,
                adjustedPreviousIndex === null ? undefined : adjustedPreviousIndex,
            ],
        ];
    }
}
//# sourceMappingURL=virtual-list-view-handler.js.map