import { ChangeDetectionStrategy, Component, ContentChild, ElementRef, Optional, Output, ViewChild, } from '@angular/core';
import { defer, ReplaySubject, Subject } from 'rxjs';
import { distinctUntilChanged, takeUntil } from 'rxjs/operators';
import { RxVirtualScrollStrategy, RxVirtualScrollViewport, RxVirtualViewRepeater, } from './model';
import { observeElementSize } from './observe-element-size';
import { getZoneUnPatchedApi } from './util';
import * as i0 from "@angular/core";
import * as i1 from "./model";
/**
 * @Component RxVirtualScrollViewport
 *
 * @description
 * Container component comparable to CdkVirtualScrollViewport acting as viewport
 * for `*rxVirtualFor` to operate on.
 *
 * Its main purpose is to implement the `RxVirtualScrollViewport` interface
 * as well as maintaining the scroll runways' height in order to give
 * the provided `RxVirtualScrollStrategy` room to position items.
 *
 * Furthermore, it will gather and forward all events to the consumer of `rxVirtualFor`.
 *
 * @docsCategory RxVirtualFor
 * @docsPage RxVirtualFor
 * @publicApi
 */
export class RxVirtualScrollViewportComponent {
    /** @internal */
    constructor(elementRef, scrollStrategy) {
        this.elementRef = elementRef;
        this.scrollStrategy = scrollStrategy;
        /** @internal */
        this.rendered$ = defer(() => this.viewRepeater.rendered$);
        /** @internal */
        this._elementScrolled = new Subject();
        this.elementScrolled$ = this._elementScrolled.asObservable();
        /** @internal */
        this._containerRect$ = new ReplaySubject(1);
        this.containerRect$ = this._containerRect$.asObservable();
        /**
         * @description
         *
         * The range to be rendered by `*rxVirtualFor`. This value is determined by the
         * provided `RxVirtualScrollStrategy`. It gives the user information about the
         * range of items being actually rendered to the DOM.
         * Note this value updates before the `renderCallback` kicks in, thus it is only
         * in sync with the DOM when the next `renderCallback` emitted an event.
         */
        this.viewRange = this.scrollStrategy.renderedRange$;
        /**
         * @description
         *
         * The index of the currently scrolled item. The scrolled item is the topmost
         * item actually being visible to the user.
         */
        this.scrolledIndexChange = this.scrollStrategy.scrolledIndex$;
        /** @internal */
        this.destroy$ = new Subject();
        this.scrollListener = (event) => this._elementScrolled.next(event);
        if (ngDevMode && !scrollStrategy) {
            throw Error('Error: rx-virtual-scroll-viewport requires an `RxVirtualScrollStrategy` to be set.');
        }
    }
    /** @internal */
    ngOnInit() {
        observeElementSize(this.elementRef.nativeElement, {
            extract: (entries) => ({
                height: Math.round(entries[0].contentRect.height),
                width: Math.round(entries[0].contentRect.width),
            }),
        })
            .pipe(distinctUntilChanged(({ height: prevHeight, width: prevWidth }, { height, width }) => prevHeight === height && prevWidth === width), takeUntil(this.destroy$))
            .subscribe(this._containerRect$);
    }
    ngAfterViewInit() {
        getZoneUnPatchedApi(this.scrollContainer(), 'addEventListener').call(this.scrollContainer(), 'scroll', this.scrollListener, {
            passive: true,
        });
    }
    /** @internal */
    ngAfterContentInit() {
        if (ngDevMode && !this.viewRepeater) {
            throw Error('Error: rx-virtual-scroll-viewport requires a `RxVirtualViewRepeater` to be provided.');
        }
        this.scrollStrategy.attach(this, this.viewRepeater);
        this.scrollStrategy.contentSize$
            .pipe(distinctUntilChanged(), takeUntil(this.destroy$))
            .subscribe((size) => this.updateContentSize(size));
    }
    /** @internal */
    ngOnDestroy() {
        getZoneUnPatchedApi(this.scrollContainer(), 'removeEventListener').call(this.scrollContainer(), 'scroll', this.scrollListener);
        this.destroy$.next();
        this.scrollStrategy.detach();
    }
    elementScrolled() {
        return this._elementScrolled.asObservable();
    }
    scrollContainer() {
        return this.scrollViewport.nativeElement;
    }
    getScrollTop() {
        return this.scrollContainer().scrollTop;
    }
    scrollTo(index, behavior) {
        // TODO: implement more complex scroll scenarios
        this.scrollContainer().scrollTo({ top: index, behavior: behavior });
    }
    scrollToIndex(index, behavior) {
        this.scrollStrategy.scrollToIndex(index, behavior);
    }
    updateContentSize(size) {
        this.runway.nativeElement.style.transform = `translate(0, ${size}px)`;
    }
}
/** @nocollapse */ RxVirtualScrollViewportComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollViewportComponent, deps: [{ token: i0.ElementRef }, { token: i1.RxVirtualScrollStrategy, optional: true }], target: i0.ɵɵFactoryTarget.Component });
/** @nocollapse */ RxVirtualScrollViewportComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "12.2.16", type: RxVirtualScrollViewportComponent, selector: "rx-virtual-scroll-viewport", outputs: { viewRange: "viewRange", scrolledIndexChange: "scrolledIndexChange" }, providers: [
        {
            provide: RxVirtualScrollViewport,
            useExisting: RxVirtualScrollViewportComponent,
        },
    ], queries: [{ propertyName: "viewRepeater", first: true, predicate: RxVirtualViewRepeater, descendants: true }], viewQueries: [{ propertyName: "runway", first: true, predicate: ["runway"], descendants: true, static: true }, { propertyName: "scrollViewport", first: true, predicate: ["scrollViewport"], descendants: true, static: true }], ngImport: i0, template: `
    <div #scrollViewport class="rx-virtual-scroll__viewport">
      <div #runway class="rx-virtual-scroll__run-way"></div>
      <ng-content></ng-content>
    </div>
  `, isInline: true, styles: ["\n      :host {\n        display: block;\n        width: 100%;\n        height: 100%;\n        box-sizing: border-box;\n        contain: strict;\n      }\n\n      :host:not([autosize]) .rx-virtual-scroll__viewport {\n        transform: translateZ(0);\n        will-change: scroll-position;\n      }\n\n      .rx-virtual-scroll__viewport {\n        contain: strict;\n        -webkit-overflow-scrolling: touch;\n        width: 100%;\n        position: absolute;\n        top: 0;\n        bottom: 0;\n        overflow: auto;\n      }\n\n      .rx-virtual-scroll__run-way {\n        width: 1px;\n        height: 1px;\n        contain: strict;\n        position: absolute;\n        will-change: transform;\n      }\n    "], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.16", ngImport: i0, type: RxVirtualScrollViewportComponent, decorators: [{
            type: Component,
            args: [{
                    selector: 'rx-virtual-scroll-viewport',
                    template: `
    <div #scrollViewport class="rx-virtual-scroll__viewport">
      <div #runway class="rx-virtual-scroll__run-way"></div>
      <ng-content></ng-content>
    </div>
  `,
                    providers: [
                        {
                            provide: RxVirtualScrollViewport,
                            useExisting: RxVirtualScrollViewportComponent,
                        },
                    ],
                    styles: [
                        `
      :host {
        display: block;
        width: 100%;
        height: 100%;
        box-sizing: border-box;
        contain: strict;
      }

      :host:not([autosize]) .rx-virtual-scroll__viewport {
        transform: translateZ(0);
        will-change: scroll-position;
      }

      .rx-virtual-scroll__viewport {
        contain: strict;
        -webkit-overflow-scrolling: touch;
        width: 100%;
        position: absolute;
        top: 0;
        bottom: 0;
        overflow: auto;
      }

      .rx-virtual-scroll__run-way {
        width: 1px;
        height: 1px;
        contain: strict;
        position: absolute;
        will-change: transform;
      }
    `,
                    ],
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }, { type: i1.RxVirtualScrollStrategy, decorators: [{
                    type: Optional
                }] }]; }, propDecorators: { runway: [{
                type: ViewChild,
                args: ['runway', { static: true }]
            }], scrollViewport: [{
                type: ViewChild,
                args: ['scrollViewport', { static: true }]
            }], viewRepeater: [{
                type: ContentChild,
                args: [RxVirtualViewRepeater]
            }], viewRange: [{
                type: Output
            }], scrolledIndexChange: [{
                type: Output
            }] } });
//# sourceMappingURL=virtual-scroll-viewport.component.js.map