import { onStrategy, strategyHandling, } from '@rx-angular/cdk/render-strategies';
import { combineLatest, concat, NEVER, of, Subject, } from 'rxjs';
import { distinctUntilChanged, ignoreElements } from 'rxjs/operators';
import { catchError, filter, map, switchMap, tap } from 'rxjs/operators';
import { getVirtualTemplateHandler } from './virtual-list-view-handler';
export function createVirtualListManager(config) {
    const { templateSettings, renderSettings, trackBy, iterableDiffers } = config;
    const { defaultStrategyName, strategies, cdRef: injectingViewCdRef, parent, patchZone, } = renderSettings;
    const errorHandler = renderSettings.errorHandler;
    const strategyHandling$ = strategyHandling(defaultStrategyName, strategies);
    let _differ;
    function getDiffer(values) {
        if (_differ) {
            return _differ;
        }
        return values
            ? (_differ = iterableDiffers.find(values).create(trackBy))
            : null;
    }
    //               type,  payload
    const listViewHandler = getVirtualTemplateHandler(Object.assign({}, templateSettings));
    const viewContainerRef = templateSettings.viewContainerRef;
    let partiallyFinished = false;
    let renderedRange;
    const _viewsRendered$ = new Subject();
    const _viewRendered$ = new Subject();
    const _renderingStart$ = new Subject();
    const viewRenderedSub = listViewHandler.viewRendered$
        .pipe(filter(({ view, change }) => !!view && change !== 1 /* remove */), map(({ view, item, index }) => ({
        view,
        item,
        index,
    })))
        .subscribe(_viewRendered$);
    function handleError() {
        return (o$) => o$.pipe(catchError((err) => {
            partiallyFinished = false;
            errorHandler === null || errorHandler === void 0 ? void 0 : errorHandler.handleError(err);
            return of(null);
        }));
    }
    return {
        viewsRendered$: _viewsRendered$,
        viewRendered$: _viewRendered$,
        renderingStart$: _renderingStart$,
        nextStrategy(nextConfig) {
            strategyHandling$.next(nextConfig);
        },
        render(values$, range$) {
            return combineLatest([
                values$.pipe(map((values) => Array.isArray(values)
                    ? values
                    : values != null
                        ? Array.from(values)
                        : [])),
                range$,
                strategyHandling$.strategy$.pipe(distinctUntilChanged()),
            ]).pipe(render(), handleError());
        },
        detach() {
            viewRenderedSub.unsubscribe();
            listViewHandler.detach();
        },
    };
    function render() {
        return (o$) => o$.pipe(
        // map iterable to latest diff
        map(([items, range, strategy]) => {
            renderedRange = range;
            const iterable = items.slice(range.start, range.end);
            const differ = getDiffer(iterable);
            let changes = null;
            if (differ) {
                if (partiallyFinished) {
                    const currentIterable = [];
                    for (let i = 0, ilen = viewContainerRef.length; i < ilen; i++) {
                        const viewRef = viewContainerRef.get(i);
                        currentIterable[i] = viewRef.context.$implicit;
                    }
                    differ.diff(currentIterable);
                    // partiallyFinished = false;
                }
                changes = differ.diff(iterable);
            }
            return {
                changes,
                itemsToRender: iterable,
                count: items.length,
                strategy,
            };
        }), 
        // Cancel old renders
        switchMap(({ changes, count, itemsToRender, strategy }) => {
            if (!changes) {
                return NEVER;
            }
            const { work$, insertedOrRemoved } = listViewHandler.toTemplateWork(changes, itemsToRender, renderedRange, strategy, count, patchZone);
            partiallyFinished = true;
            const notifyParent = insertedOrRemoved && parent;
            _renderingStart$.next();
            return combineLatest(
            // emit after all changes are rendered
            work$.length > 0 ? work$ : [of(itemsToRender)]).pipe(tap(() => {
                partiallyFinished = false;
                const viewsRendered = [];
                const end = viewContainerRef.length;
                let i = 0;
                for (i; i < end; i++) {
                    viewsRendered.push(viewContainerRef.get(i));
                }
                _viewsRendered$.next(viewsRendered);
            }), notifyParent
                ? switchMap((v) => concat(of(v), onStrategy(injectingViewCdRef, strategy, (_v, work, options) => {
                    work(injectingViewCdRef, options.scope);
                }, {
                    ngZone: patchZone,
                    scope: injectingViewCdRef.context ||
                        injectingViewCdRef,
                }).pipe(ignoreElements())))
                : (o$) => o$, handleError(), map(() => itemsToRender));
        }));
    }
}
//# sourceMappingURL=virtual-template-manager.js.map